# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import json
import logging
import sys
from dataclasses import dataclass
from importlib.metadata import version as get_package_version

import requests

from atopile.config import config
from faebryk.libs.picker.api.models import (
    BaseParams,
    Component,
    LCSCParams,
    ManufacturerPartParams,
)
from faebryk.libs.util import ConfigFlag, once

logger = logging.getLogger(__name__)

DEFAULT_API_TIMEOUT_SECONDS = 30

API_LOG = ConfigFlag("API_LOG", descr="Log API calls (very verbose)", default=False)


class ApiError(Exception): ...


class ApiNotConfiguredError(ApiError): ...


class ApiHTTPError(ApiError):
    def __init__(self, error: requests.exceptions.HTTPError):
        super().__init__()
        self.response = error.response

    def __str__(self) -> str:
        status_code = self.response.status_code
        try:
            detail = self.response.json()["detail"]
        except Exception:
            detail = self.response.text
        return f"{super().__str__()}: {status_code} {detail}"


class ApiClient:
    @dataclass
    class ApiConfig:
        api_url: str = config.project.services.components.url
        api_key: str | None = None

    def __init__(self):
        self._client = requests.Session()
        self._client.headers.update(
            {
                "User-Agent": (
                    f"atopile/{get_package_version('atopile')} "
                    f"({sys.platform}; "
                    f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro})"
                ),
            }
        )

    @property
    @once
    def _cfg(self) -> ApiConfig:
        return self.ApiConfig()

    @property
    @once
    def _headers(self) -> dict[str, str]:
        return {"Authorization": f"Bearer {self._cfg.api_key}"}

    def _get(self, url: str, timeout: float = 10) -> requests.Response:
        try:
            response = self._client.get(
                f"{self._cfg.api_url}{url}", timeout=timeout, headers=self._headers
            )
            response.raise_for_status()
        except requests.exceptions.HTTPError as e:
            raise ApiHTTPError(e) from e

        if API_LOG:
            logger.debug(
                "GET %s%s\n->\n%s",
                self._cfg.api_url,
                url,
                json.dumps(response.json(), indent=2),
            )
        else:
            logger.debug("GET %s%s", self._cfg.api_url, url)

        return response

    def _post(
        self, url: str, data: dict, timeout: float = DEFAULT_API_TIMEOUT_SECONDS
    ) -> requests.Response:
        try:
            response = self._client.post(
                f"{self._cfg.api_url}{url}",
                json=data,
                timeout=timeout,
                headers=self._headers,
            )
            response.raise_for_status()
        except requests.exceptions.HTTPError as e:
            raise ApiHTTPError(e) from e

        if API_LOG:
            logger.debug(
                "POST %s%s\n%s\n->\n%s",
                self._cfg.api_url,
                url,
                json.dumps(data, indent=2),
                json.dumps(response.json(), indent=2),
            )
        else:
            logger.debug(
                "POST %s%s\n%s", self._cfg.api_url, url, json.dumps(data, indent=2)
            )

        return response

    @once
    def fetch_part_by_lcsc(self, lcsc: int) -> list["Component"]:
        response = self._get(f"/v0/component/lcsc/{lcsc}")
        return [Component.from_dict(part) for part in response.json()["components"]]  # type: ignore

    @once
    def fetch_part_by_mfr(self, mfr: str, mfr_pn: str) -> list["Component"]:
        response = self._get(f"/v0/component/mfr/{mfr}/{mfr_pn}")
        return [Component.from_dict(part) for part in response.json()["components"]]  # type: ignore

    def query_parts(self, method: str, params: BaseParams) -> list["Component"]:
        response = self._post(f"/v0/query/{method}", params.serialize())
        return [Component.from_dict(part) for part in response.json()["components"]]  # type: ignore

    @once
    def fetch_parts(self, params: BaseParams) -> list["Component"]:
        assert params.endpoint
        return self.query_parts(params.endpoint, params)

    def fetch_parts_multiple(
        self, params: list[BaseParams | LCSCParams | ManufacturerPartParams]
    ) -> list[list["Component"]]:
        # TODO: batch queries
        response = self._post("/v0/query", {"queries": [p.serialize() for p in params]})
        results = [
            [Component.from_dict(part) for part in result["components"]]  # type: ignore
            for result in response.json()["results"]
        ]

        if len(results) != len(params):
            raise ApiError(f"Expected {len(params)} results, got {len(results)}")

        return results


@once
def get_api_client() -> ApiClient:
    return ApiClient()



import logging
from pathlib import Path

from atopile.mcp.util import (
    Language,
    MCPTools,
    NodeInfo,
    NodeInfoOverview,
    NodeType,
)
from faebryk.core.module import Module
from faebryk.core.moduleinterface import ModuleInterface
from faebryk.core.node import Node

logger = logging.getLogger(__name__)

library_tools = MCPTools()


def _get_library_nodes(
    t: type[Node] | tuple[type[Node], ...],
) -> list[NodeInfoOverview]:
    import faebryk.library._F as F

    return [
        NodeInfoOverview(
            name=m.__name__,
            docstring=m.__doc__ or "",
            language=Language.FABLL,
            type=NodeType.MODULE if issubclass(m, Module) else NodeType.INTERFACE,
            inherits=m.__bases__[0].__name__
            if m.__bases__[0] not in [Module, ModuleInterface]
            else None,
        )
        for m in F.__dict__.values()
        if isinstance(m, type) and issubclass(m, t)
    ]


def _locator_from_file_path(path: Path) -> str:
    return f"#file://{path.as_posix()}"


def _get_library_node(name: str, t: type[Node] = Node) -> NodeInfo:
    import faebryk.library._F as F

    if name not in F.__dict__:
        raise ValueError(f"Type {name} not found")

    m = F.__dict__[name]

    if not isinstance(m, type) or not issubclass(m, t):
        raise ValueError(f"Type {name} is not a valid {t.__name__}")

    # get file of module
    file = m.__module__
    if file is None:
        raise ValueError(f"Type {name} is not part of a module")

    # import the module to get its file path
    import importlib

    module = importlib.import_module(file)

    if module.__file__ is None:
        raise ValueError(f"Type {name} has no file")

    # read file
    filepath = module.__file__
    path = Path(filepath)
    code = path.read_text(encoding="utf-8")

    node = getattr(module, name)
    # get docstring
    docstring = node.__doc__ or ""

    language = Language.FABLL
    locator = _locator_from_file_path(path)
    return NodeInfo(
        name=name, docstring=docstring, locator=locator, language=language, code=code
    )


@library_tools.register()
def inspect_library_module_or_interface(name: str) -> NodeInfo:
    """
    Inspect a standard library module or interface, returning information about its
    name, docstring, locator, language, and code.
    """
    return _get_library_node(name)


@library_tools.register()
def get_library_modules_or_interfaces(
    include_modules: bool = True, include_interfaces: bool = True
) -> list[NodeInfoOverview]:
    """
    List all atopile standard library modules and interfaces.
    """
    types = tuple()
    if include_modules:
        types += (Module,)
    if include_interfaces:
        types += (ModuleInterface,)
    return _get_library_nodes(types)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT


import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.core.moduleinterface import ModuleInterface


class has_single_electric_reference_shared(F.has_single_electric_reference.impl()):
    lazy: F.is_lazy

    def __init__(self, gnd_only: bool = False):
        super().__init__()
        self.gnd_only = gnd_only

    def on_obj_set(self):
        super().on_obj_set()

        if not isinstance(self.obj, (Module, ModuleInterface)):
            raise TypeError(
                f"has_single_electric_reference_shared can only be used on "
                f"modules or module interfaces, got {self.obj}"
            )

        F.ElectricSignal.connect_all_module_references(self.obj, gnd_only=self.gnd_only)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
from dataclasses import dataclass, field
from pathlib import Path
from textwrap import indent

from atopile.front_end import _FeatureFlags

logger = logging.getLogger(__name__)


class _StrBuilder:
    def __init__(self) -> None:
        self.out = ""

    def append(self, content: str) -> None:
        self.out += content

    def append_line(self, line: str) -> None:
        self.append(line + "\n")

    def append_indented(self, content: str) -> None:
        self.append_line(indent(content, " " * 4))

    def spacer(self) -> None:
        if self.out:
            self.append_line("")

    def dump(self) -> str:
        return self.out


class AtoCodeGen:
    """
    This is the ghetto version of my other ato code generator that is not upstream.
    That one is a lot more complex, but in features and usage.
    So for the second we stick with this boi.
    """

    @dataclass(frozen=True, eq=True)
    class Import:
        name: str
        path: Path | None = None

    Experiment = _FeatureFlags.Feature

    @dataclass(kw_only=True)
    class Statement:
        _commented: bool = False

        def dump(self) -> str:
            out = _StrBuilder()
            if self._commented:
                out.append("#")
            out.append(self.dump_stmt())
            return out.dump()

        def dump_stmt(self) -> str: ...

        def comment_out(self):
            self._commented = True

    @dataclass
    class Trait(Statement):
        name: str
        args: dict[str, str] = field(default_factory=dict)
        constructor: str | None = None

        def dump_stmt(self) -> str:
            out = f"trait {self.name}"
            if self.constructor:
                out += f"::{self.constructor}"
            if self.args:
                out += f"<{', '.join(f'{k}="{v}"' for k, v in self.args.items())}>"
            return out

    @dataclass
    class PinDeclaration(Statement):
        name: str

        def dump_stmt(self) -> str:
            return f"pin {self.name}"

    @dataclass
    class Connect(Statement):
        class Connectable:
            def __init__(self, name: str, declare: str | None = None) -> None:
                self.name = name
                self.declare = declare

            def dump(self) -> str:
                out = _StrBuilder()
                if self.declare:
                    out.append(f"{self.declare} ")
                out.append(self.name)
                return out.dump()

        left: Connectable
        right: Connectable

        def dump_stmt(self) -> str:
            return f"{self.left.dump()} ~ {self.right.dump()}"

    class Spacer(Statement):
        def dump(self) -> str:
            return ""

    @dataclass
    class Comment(Statement):
        text: str

        def dump_stmt(self) -> str:
            return f"# {self.text}"

        @classmethod
        def from_lines(cls, *lines: str) -> list["AtoCodeGen.Comment"]:
            return [cls(line) for line in lines]

    @dataclass
    class ComponentFile:
        identifier: str | None = None
        imports: set["AtoCodeGen.Import"] = field(default_factory=set)
        experiments: set["AtoCodeGen.Experiment"] = field(default_factory=set)
        stmts: list["AtoCodeGen.Statement"] = field(default_factory=list)
        docstring: str | None = None

        def dump(self) -> str:
            if self.identifier is None:
                raise ValueError("identifier is required")

            out = _StrBuilder()

            for exp in self.experiments:
                out.append_line(f'#pragma experiment("{exp.value}")')

            for imp in sorted(self.imports, key=lambda x: x.name):
                if imp.path:
                    out.append(f"from {imp.path} ")
                out.append_line(f"import {imp.name}")

            out.spacer()

            out.append_line(f"component {self.identifier}:")
            if self.docstring:
                out.append_indented(f'"""{self.docstring}"""')
                out.spacer()

            for stmt in self.stmts:
                out.append_indented(stmt.dump())

            return out.dump()

        def add_stmt(self, stmt: "AtoCodeGen.Statement") -> None:
            self.stmts.append(stmt)

        def add_comments(self, *comments: str, use_spacer: bool = False) -> None:
            self.add_stmts(
                *AtoCodeGen.Comment.from_lines(*comments), use_spacer=use_spacer
            )

        def add_trait(
            self,
            name: str,
            constructor: str | None = None,
            auto_import: bool = True,
            **args: str | None,
        ) -> "AtoCodeGen.Trait":
            self.enable_experiment(AtoCodeGen.Experiment.TRAITS)

            if auto_import:
                self.imports.add(AtoCodeGen.Import(name))

            trait = AtoCodeGen.Trait(
                name,
                args={k: v for k, v in args.items() if v is not None},
                constructor=constructor,
            )
            self.add_stmt(trait)
            return trait

        def enable_experiment(self, experiment: "AtoCodeGen.Experiment") -> None:
            self.experiments.add(experiment)

        def add_stmts(
            self, *stmts: "AtoCodeGen.Statement", use_spacer: bool = True
        ) -> None:
            if use_spacer:
                self.add_stmt(AtoCodeGen.Spacer())
            for stmt in stmts:
                self.add_stmt(stmt)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
from enum import StrEnum

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.iso_metric_screw_thread import Iso262_MetricScrewThreadSizes
from faebryk.libs.library import L

logger = logging.getLogger(__name__)


class Mounting_Hole(Module):
    class PadType(StrEnum):
        NoPad = ""
        Pad = "Pad"
        Pad_TopBottom = "Pad_TopBottom"
        Pad_TopOnly = "Pad_TopOnly"
        Pad_Via = "Pad_Via"

    contact: F.Electrical

    attach_to_footprint: F.can_attach_to_footprint_symmetrically
    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.H
    )

    @L.rt_field
    def footprint(self):
        size_mm = f"{self._diameter.value:.1f}mm"
        size_name = self._diameter.name.replace("_", ".")
        padtype = self._pad_type

        if size_name:
            size_name = f"_{size_name}"
        if padtype:
            padtype = f"_{padtype}"

        fp = F.KicadFootprint(pin_names=["1"])
        fp.add(
            F.KicadFootprint.has_kicad_identifier(
                f"MountingHole:MountingHole_{size_mm}{size_name}{padtype}"
            )
        )
        return F.has_footprint_defined(fp)

    def __init__(self, diameter: Iso262_MetricScrewThreadSizes, pad_type: PadType):
        super().__init__()
        self._diameter = diameter
        self._pad_type = pad_type

    def __preinit__(self):
        if self._pad_type is not self.PadType.NoPad:
            # footprint has 1 contact
            self.footprint.get_footprint().get_trait(F.can_attach_via_pinmap).attach(
                pinmap={"1": self.contact}
            )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
from pathlib import Path

import freetype
from shapely import Point, Polygon

from faebryk.libs.geometry.basic import (
    flatten_polygons,
    transform_polygon,
)

logger = logging.getLogger(__name__)


class Font:
    def __init__(self, ttf: Path):
        super().__init__()

        self.path = ttf

    def string_to_polygons(
        self,
        string: str,
        font_size: float,
        bbox: tuple[float, float] | None = None,
        wrap: bool = False,
        scale_to_fit: bool = False,
    ) -> list[Polygon]:
        """
        Render the polygons of a string from a ttf font file

        :param ttf_path: Path to the ttf font file
        :param string: The string to extract
        :param font_size: The font size in points
        :param bbox: The bounding box to fit the text in, in points
        :param wrap: Wrap the text to fit the bounding box
        :param scale_to_fit: Scale the text to fit the bounding box
        :return: A list of polygons that represent the string
        """

        if wrap and not bbox:
            raise ValueError("Bounding box must be given when wrapping text")

        if scale_to_fit and not bbox:
            raise ValueError("Bounding box must be given when fitting text")

        if wrap and scale_to_fit:
            raise NotImplementedError("Cannot wrap and scale to fit at the same time")

        # TODO: use bezier control points in outline.tags

        face = freetype.Face(str(self.path))
        polygons = []
        offset = Point(0, 0)

        if scale_to_fit:
            font_size = 1

        text_size = Point(0, 0)

        scale = font_size / face.units_per_EM
        for char in string:
            face.load_char(char)

            if bbox and not scale_to_fit:
                if offset.x + face.glyph.advance.x > bbox[0] / scale:
                    if not wrap:
                        break
                    offset = Point(0, offset.y + face.glyph.advance.y)
                    if offset.y > bbox[1] / scale:
                        break

            points = face.glyph.outline.points
            contours = face.glyph.outline.contours

            start = 0

            for contour in contours:
                contour_points = [Point(p) for p in points[start : contour + 1]]
                contour_points.append(contour_points[0])
                start = contour + 1
                contour_points = [
                    Point(p.x + offset.x, p.y + offset.y) for p in contour_points
                ]
                polygons.append(Polygon(contour_points))

            offset = Point(offset.x + face.glyph.advance.x, offset.y)

            if not wrap or not bbox:
                continue

            if offset.x > bbox[0]:
                offset = Point(0, offset.y + face.glyph.advance.y)
                if offset.y > bbox[1]:
                    break

        bounds = [p.bounds for p in polygons]
        min_x, min_y, max_x, max_y = (
            min(b[0] for b in bounds),
            min(b[1] for b in bounds),
            max(b[2] for b in bounds),
            max(b[3] for b in bounds),
        )
        offset = Point(
            -min_x,
            -min_y,
        )

        if scale_to_fit and bbox:
            scale = min(bbox[0] / (max_x - min_x), bbox[1] / (max_y - min_y))

        logger.debug(f"Text size: {text_size}")
        logger.debug(f"Offset: {offset}")
        logger.debug(f"Scale: {scale}")

        polygons = flatten_polygons(polygons)
        polygons = [
            transform_polygon(p, scale=scale, offset=(offset.x, offset.y))
            for p in polygons
        ]

        # Invert the y-axis
        max_y = max(p.bounds[3] for p in polygons)
        polygons = [
            Polygon([(p[0], -p[1] + max_y) for p in polygon.exterior.coords])
            for polygon in polygons
        ]

        return polygons



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

from enum import Enum, auto

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.core.node import rt_field
from faebryk.libs.library import L


class BJT(Module):
    class DopingType(Enum):
        NPN = auto()
        PNP = auto()

    # TODO use this, here is more info: https://en.wikipedia.org/wiki/Bipolar_junction_transistor#Regions_of_operation
    class OperationRegion(Enum):
        ACTIVE = auto()
        INVERTED = auto()
        SATURATION = auto()
        CUT_OFF = auto()

    doping_type = L.p_field(domain=L.Domains.ENUM(DopingType))
    operation_region = L.p_field(domain=L.Domains.ENUM(OperationRegion))

    emitter: F.Electrical
    base: F.Electrical
    collector: F.Electrical

    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.Q
    )

    @rt_field
    def can_bridge(self):
        return F.can_bridge_defined(self.collector, self.emitter)

    @rt_field
    def pin_association_heuristic(self):
        return F.has_pin_association_heuristic_lookup_table(
            mapping={
                self.emitter: ["E", "Emitter"],
                self.base: ["B", "Base"],
                self.collector: ["C", "Collector"],
            },
            accept_prefix=False,
            case_sensitive=False,
        )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
import sys
from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum, auto
from importlib.metadata import version as get_package_version
from pathlib import Path
from typing import Any, Literal
from urllib.parse import urlparse

import requests
from dataclasses_json import config as dataclasses_json_config
from dataclasses_json.api import dataclass_json

from atopile.config import config
from faebryk.libs.package.artifacts import Artifacts
from faebryk.libs.package.dist import Dist
from faebryk.libs.util import indented_container, once

logger = logging.getLogger(__name__)


_EXCLUDE_FROM_JSON = dataclasses_json_config(exclude=lambda _: True)


class _Type(Enum):
    GET = auto()
    POST = auto()


class _Schemas:
    @dataclass_json
    @dataclass(frozen=True)
    class Author:
        name: str
        email: str

    @dataclass_json
    @dataclass(frozen=True)
    class PackageStats:
        total_downloads: int | None
        this_week_downloads: int | None
        this_month_downloads: int | None

    @dataclass_json
    @dataclass(frozen=True)
    class FileHashes:
        sha256: str

    @dataclass_json
    @dataclass(frozen=True)
    class RegistryDependencySpec:
        identifier: str
        type: Literal["registry"] = "registry"
        release: str | None = None

    @dataclass_json
    @dataclass(frozen=True)
    class PackageDependencies:
        requires: list["_Schemas.RegistryDependencySpec"]

    @dataclass_json
    @dataclass(frozen=True)
    class ArtifactInfo:
        filename: str
        url: str
        size: int
        hashes: "_Schemas.FileHashes"
        build_name: str | None = None

    @dataclass_json
    @dataclass(frozen=True)
    class ArtifactsInfo:
        artifacts: list["_Schemas.ArtifactInfo"]

    @dataclass_json
    @dataclass(frozen=True)
    class LayoutInfo:
        build_name: str
        url: str

    @dataclass_json
    @dataclass(frozen=True)
    class LayoutsInfo:
        layouts: list["_Schemas.LayoutInfo"]

    @dataclass_json
    @dataclass(frozen=True)
    class PackageReleaseInfo:
        created_at: datetime = field(
            metadata=dataclasses_json_config(decoder=datetime.fromisoformat)
        )
        released_at: datetime = field(
            metadata=dataclasses_json_config(decoder=datetime.fromisoformat)
        )
        key: str = field(metadata=_EXCLUDE_FROM_JSON)
        identifier: str
        version: str
        repository: str
        authors: list["_Schemas.Author"]
        license: str
        summary: str
        homepage: str | None
        readme_url: str | None = field(metadata=_EXCLUDE_FROM_JSON)
        url: str
        stats: "_Schemas.PackageStats"
        hashes: "_Schemas.FileHashes"
        filename: str
        git_ref: str | None
        requires_atopile: str
        size: int
        download_url: str = field(metadata=_EXCLUDE_FROM_JSON)
        builds: list[str] | None
        dependencies: "_Schemas.PackageDependencies"
        artifacts: "_Schemas.ArtifactsInfo | None"
        layouts: "_Schemas.LayoutsInfo | None"
        yanked_at: str | None
        yanked_reason: str | None

    @dataclass_json
    @dataclass(frozen=True)
    class Package:
        info: "_Schemas.PackageInfo"
        readme: str | None

    @dataclass_json
    @dataclass(frozen=True)
    class PackageRelease:
        info: "_Schemas.PackageReleaseInfo"
        readme: str | None

    @dataclass_json
    @dataclass(frozen=True)
    class PackageInfo:
        created_at: datetime | str
        released_at: datetime | str
        key: str
        identifier: str
        version: str
        repository: str
        authors: list["_Schemas.Author"]
        license: str
        summary: str
        homepage: str | None
        readme_url: str | None
        url: str
        stats: "_Schemas.PackageStats"

    @dataclass_json
    @dataclass(frozen=True)
    class PackageInfoBrief:
        identifier: str
        version: str
        summary: str
        url: str
        repository: str
        homepage: str | None

    @dataclass_json
    @dataclass(frozen=True)
    class QueryResult:
        packages: list["_Schemas.PackageInfoBrief"]

    @dataclass_json
    @dataclass(frozen=True)
    class PresignedUploadInfo:
        url: str
        fields: dict[str, str]

    @dataclass_json
    @dataclass(frozen=True)
    class PublishRequest:
        identifier: str
        package_version: str
        git_ref: str | None
        package_size: int
        artifacts_size: int | None
        manifest: dict[str, Any]

    @dataclass_json
    @dataclass(frozen=True)
    class PublishRequestResponse:
        release_id: str
        upload_info: "_Schemas.PresignedUploadInfo"
        artifacts_upload_info: "_Schemas.PresignedUploadInfo"
        status: Literal["ok"] = "ok"

    @dataclass_json
    @dataclass(frozen=True)
    class PublishCompletedResponse:
        url: str
        status: Literal["ok"] = "ok"

    @dataclass_json
    @dataclass(frozen=True)
    class UploadCompleteRequest:
        release_id: str

    @dataclass_json
    @dataclass(frozen=True)
    class ValidationError:
        loc: list[str | int]
        msg: str
        type: str

    @dataclass_json
    @dataclass(frozen=True)
    class HTTPValidationError:
        detail: list["_Schemas.ValidationError"]

    @dataclass_json
    @dataclass(frozen=True)
    class PackageReleases:
        releases: list["_Schemas.PackageReleaseInfo"]


class _Endpoints:
    class PackageReleases:
        TYPE = _Type.GET

        @staticmethod
        def url(request: "_Endpoints.PackageReleases.Request") -> str:
            return f"/v1/package/{request.identifier}/releases"

        @dataclass_json
        @dataclass(frozen=True)
        class Request:
            identifier: str

        Response = _Schemas.PackageReleases

    class PackageRelease:
        TYPE = _Type.GET

        @staticmethod
        def url(request: "_Endpoints.PackageRelease.Request") -> str:
            return f"/v1/package/{request.identifier}/releases/{request.version}"

        @dataclass_json
        @dataclass(frozen=True)
        class Request:
            identifier: str
            version: str

        Response = _Schemas.PackageRelease

    class Package:
        TYPE = _Type.GET

        @staticmethod
        def url(request: "_Endpoints.Package.Request") -> str:
            return f"/v1/package/{request.identifier}"

        @dataclass_json
        @dataclass(frozen=True)
        class Request:
            identifier: str

        Response = _Schemas.Package

    class Packages:
        TYPE = _Type.GET

        @staticmethod
        def url(request: "_Endpoints.Packages.Request") -> str:
            return f"/v1/packages?query={request.query}"

        @dataclass_json
        @dataclass(frozen=True)
        class Request:
            query: str

        Response = _Schemas.QueryResult

    class Publish:
        TYPE = _Type.POST

        @staticmethod
        def url() -> str:
            return "/v1/publish"

        Request = _Schemas.PublishRequest
        Response = _Schemas.PublishRequestResponse

    class PublishUploadComplete:
        TYPE = _Type.POST

        @staticmethod
        def url() -> str:
            return "/v1/publish/upload-complete"

        Request = _Schemas.UploadCompleteRequest
        Response = _Schemas.PublishCompletedResponse


class Errors:
    class PackagesApiError(Exception): ...

    class PackagesApiHTTPError(Exception):
        def __init__(self, error: requests.exceptions.HTTPError, detail: str):
            super().__init__()
            self.error = error
            self.response = error.response
            self.detail = detail

        @classmethod
        def from_http(cls, error: "Errors.PackagesApiHTTPError"):
            return cls(error.error, error.detail)

        @property
        def code(self) -> int:
            return self.response.status_code

        def __str__(self) -> str:
            return f"{type(self).__name__}: {self.code} {self.detail}"

    class AuthenticationError(PackagesApiError): ...

    class PackageNotFoundError(PackagesApiHTTPError):
        def __init__(
            self,
            error: requests.exceptions.HTTPError,
            detail: str,
            package_identifier: str,
        ):
            super().__init__(error, detail)
            self.package_identifier = package_identifier

        def __str__(self) -> str:
            return f"{type(self).__name__}: {self.package_identifier}: {self.detail}"

        @classmethod
        def from_http(  # type: ignore
            cls, error: "Errors.PackagesApiHTTPError", package_identifier: str
        ):
            return cls(error.error, error.detail, package_identifier)

    class InvalidPackageIdentifierError(PackagesApiHTTPError):
        def __init__(
            self,
            error: requests.exceptions.HTTPError,
            detail: str,
            package_identifier: str,
        ):
            super().__init__(error, detail)
            self.package_identifier = package_identifier

        def __str__(self) -> str:
            return f"{type(self).__name__}: {self.package_identifier}: {self.detail}"

        @classmethod
        def from_http(  # type: ignore
            cls, error: "Errors.PackagesApiHTTPError", package_identifier: str
        ):
            return cls(error.error, error.detail, package_identifier)

    class ReleaseNotFoundError(PackagesApiHTTPError):
        def __init__(
            self,
            error: requests.exceptions.HTTPError,
            detail: str,
            package_identifier: str,
            release: str,
        ):
            super().__init__(error, detail)
            self.package_identifier = package_identifier
            self.release = release

        def __str__(self) -> str:
            return (
                f"{type(self).__name__}: {self.package_identifier}@{self.release}:"
                f" {self.detail}"
            )

        @classmethod
        def from_http(  # type: ignore
            cls,
            error: "Errors.PackagesApiHTTPError",
            package_identifier: str,
            release: str,
        ):
            return cls(error.error, error.detail, package_identifier, release)

    class ReleaseAlreadyExistsError(PackagesApiHTTPError): ...


class PackagesAPIClient:
    @dataclass
    class ApiConfig:
        api_url: str = config.project.services.packages.url

    @property
    @once
    def _cfg(self) -> ApiConfig:
        return self.ApiConfig()

    def __init__(self):
        self._client = requests.Session()
        self._client.headers.update(
            {
                "User-Agent": (
                    f"atopile/{get_package_version('atopile')} "
                    f"({sys.platform}; "
                    f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro})"
                ),
            }
        )

    def _get(
        self,
        url: str,
        timeout: float = 10,
    ) -> requests.Response:
        response = self._client.get(f"{self._cfg.api_url}{url}", timeout=timeout)
        try:
            response.raise_for_status()
        except requests.exceptions.HTTPError as e:
            try:
                detail = response.json()["detail"]
            except (requests.JSONDecodeError, KeyError):
                detail = response.text
            raise Errors.PackagesApiHTTPError(e, detail) from e
        return response

    def _post(
        self,
        url: str,
        data: Any,
        timeout: float = 10,
        authenticate: bool = False,
    ) -> requests.Response:
        headers = {}
        if authenticate:
            headers |= self._authenticate()

        response = self._client.post(
            f"{self._cfg.api_url}{url}",
            json=data,
            timeout=timeout,
            headers=headers,
        )
        try:
            response.raise_for_status()
            assert response.json()["status"] == "ok"
        except requests.exceptions.HTTPError as e:
            try:
                detail = response.json()["detail"]
            except (requests.JSONDecodeError, KeyError):
                detail = response.text
            raise Errors.PackagesApiHTTPError(e, detail) from e
        return response

    def _upload(
        self,
        url: str,
        file_path: Path,
        data: dict[str, str],
        timeout: float = 10,
        skip_verify: bool = False,
    ) -> requests.Response:
        response = self._client.post(
            url,
            data=data,
            files={
                "file": (file_path.name, file_path.read_bytes()),
                "Content-Type": "application/zip",
            },
            timeout=timeout,
            verify=not skip_verify,
        )
        try:
            response.raise_for_status()
        except requests.exceptions.HTTPError as e:
            try:
                detail = response.json()["detail"]
            except (requests.JSONDecodeError, KeyError):
                detail = response.text
            raise Errors.PackagesApiHTTPError(e, detail) from e
        return response

    def _authenticate(self) -> dict[str, str]:
        """
        Currently, the only supported authentication method is Github Actions OIDC.
        """
        try:
            from github_oidc.client import get_actions_header

            return get_actions_header(urlparse(self._cfg.api_url).netloc)
        except Exception as e:
            raise Errors.AuthenticationError(e) from e

    def publish(
        self,
        identifier: str,
        version: str,
        git_ref: str | None,
        dist: Dist,
        artifacts: Artifacts | None,
        skip_auth: bool = False,
    ) -> _Endpoints.PublishUploadComplete.Response:
        """
        Publish a package to the package registry.

        Only works in Github Actions.
        """
        ## Request upload
        try:
            r = self._post(
                _Endpoints.Publish.url(),
                data=_Endpoints.Publish.Request(
                    identifier=identifier,
                    package_version=version,
                    git_ref=git_ref,
                    package_size=dist.bytes,
                    artifacts_size=artifacts.bytes if artifacts else None,
                    manifest=dist.manifest.model_dump(mode="json"),
                ).to_dict(),  # type: ignore
                authenticate=not skip_auth,
            )
        except Errors.PackagesApiHTTPError as e:
            if e.code == 409:
                raise Errors.ReleaseAlreadyExistsError.from_http(e) from e
            raise

        response = _Endpoints.Publish.Response.from_dict(r.json())  # type: ignore

        ## Upload the package
        try:
            r = self._upload(
                url=response.upload_info.url,
                file_path=dist.path,
                data=response.upload_info.fields,
                skip_verify=skip_auth,
            )
        except Errors.PackagesApiHTTPError:
            raise

        ## Upload the artifacts
        if artifacts:
            try:
                r = self._upload(
                    url=response.artifacts_upload_info.url,
                    file_path=artifacts.path,
                    data=response.artifacts_upload_info.fields,
                    skip_verify=skip_auth,
                )
            except Errors.PackagesApiHTTPError:
                raise

        ## Confirm upload
        try:
            r = self._post(
                _Endpoints.PublishUploadComplete.url(),
                data=_Endpoints.PublishUploadComplete.Request(
                    release_id=response.release_id,
                ).to_dict(),  # type: ignore
                authenticate=not skip_auth,
            )
        except Errors.PackagesApiHTTPError:
            raise

        response = _Endpoints.PublishUploadComplete.Response.from_dict(r.json())  # type: ignore
        return response

    def get_package(
        self, identifier: str, version: str | None = None
    ) -> _Endpoints.PackageRelease.Response:
        """
        Get a package from the package registry.
        """
        if version is None:
            try:
                r = self._get(
                    _Endpoints.Package.url(_Endpoints.Package.Request(identifier))
                )
            except Errors.PackagesApiHTTPError as e:
                if e.code == 404:
                    raise Errors.PackageNotFoundError.from_http(e, identifier) from e
                if e.code == 422:
                    raise Errors.InvalidPackageIdentifierError.from_http(
                        e, identifier
                    ) from e
                raise
            response = _Endpoints.Package.Response.from_dict(r.json())  # type: ignore
            assert isinstance(response, _Endpoints.Package.Response)
            version = response.info.version

        try:
            r = self._get(
                _Endpoints.PackageRelease.url(
                    _Endpoints.PackageRelease.Request(identifier, version)
                )
            )
        except Errors.PackagesApiHTTPError as e:
            if e.code == 404:
                raise Errors.ReleaseNotFoundError.from_http(
                    e, identifier, version
                ) from e
            raise
        try:
            response = _Endpoints.PackageRelease.Response.from_dict(r.json())  # type: ignore
        except Exception:
            logger.error(indented_container(r.json(), recursive=True))
            raise
        return response

    def get_release_dist(
        self, identifier: str, output_path: Path, version: str | None = None
    ) -> Dist:
        release = self.get_package(identifier, version)
        url = release.info.download_url
        filepath = output_path / release.info.filename
        filepath.parent.mkdir(parents=True, exist_ok=True)
        # use requests to download the file to output_path
        with requests.get(url, stream=True) as r:
            r.raise_for_status()
            with filepath.open("wb") as f:
                for chunk in r.iter_content(chunk_size=8192):
                    f.write(chunk)

        return Dist(filepath)

    def query_packages(self, query: str) -> _Endpoints.Packages.Response:
        r = self._get(_Endpoints.Packages.url(_Endpoints.Packages.Request(query)))
        return _Endpoints.Packages.Response.from_dict(r.json())  # type: ignore



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
from pathlib import Path

import requests

import faebryk.library._F as F
from faebryk.core.module import Module

logger = logging.getLogger(__name__)


def export_datasheets(
    app: Module,
    path: Path = Path("build/documentation/datasheets"),
    overwrite: bool = False,
):
    """
    Export all datasheets of all modules (that have a datasheet defined)
    of the given application.
    """
    # Create directories if they don't exist
    path.mkdir(parents=True, exist_ok=True)

    logger.info(f"Exporting datasheets to: {path}")
    for m in app.get_children_modules(types=Module):
        if not m.has_trait(F.has_datasheet):
            continue
        url = m.get_trait(F.has_datasheet).get_datasheet()
        if not url:
            logger.warning(f"Missing datasheet URL for {m}")
            continue
        filename = type(m).__name__ + ".pdf"
        file_path = path / filename
        if file_path.exists() and not overwrite:
            logger.debug(
                f"Datasheet for {m} already exists, skipping download"  # noqa: E501
            )
            continue
        try:
            _download_datasheet(url, file_path)
        except DatasheetDownloadException as e:
            logger.error(f"Failed to download datasheet for {m}: {e}")

        logger.debug(f"Downloaded datasheet for {m}")


class DatasheetDownloadException(Exception):
    pass


def _download_datasheet(url: str, path: Path):
    """
    Download the datasheet of the given module and save it to the given path.
    """
    if not url.endswith(".pdf"):
        raise DatasheetDownloadException(f"Datasheet URL {url} is probably not a PDF")
    if not url.startswith(("http://", "https://")):
        raise DatasheetDownloadException(
            f"Datasheet URL {url} is probably not a valid URL"
        )

    try:
        # TODO probably need something fancier
        user_agent_headers = {
            "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_11_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/50.0.2661.102 Safari/537.36"  # noqa: E501
        }
        response = requests.get(url, headers=user_agent_headers)
        response.raise_for_status()
    except requests.RequestException as e:
        raise DatasheetDownloadException(
            f"Failed to download datasheet from {url}: {e}"
        ) from e

    # check if content is pdf
    if not response.content.startswith(b"%PDF"):
        raise DatasheetDownloadException(
            f"Downloaded content is not a PDF: {response.content[:100]}"
        )

    try:
        path.write_bytes(response.content)
    except Exception as e:
        raise DatasheetDownloadException(
            f"Failed to save datasheet to {path}: {e}"
        ) from e



from unittest.mock import Mock

import pytest

from faebryk.core.node import Node


@pytest.fixture
def node_hierarchy():
    """Creates a mock node hierarchy for testing."""
    root = Mock(spec=Node)
    child1 = Mock(spec=Node)
    child2 = Mock(spec=Node)
    grandchild1 = Mock(spec=Node)
    grandchild2 = Mock(spec=Node)

    # Setup hierarchies
    root.get_hierarchy = lambda: [(root, "root")]
    child1.get_hierarchy = lambda: [(root, "root"), (child1, "child1")]
    child2.get_hierarchy = lambda: [(root, "root"), (child2, "child2")]
    grandchild1.get_hierarchy = lambda: [
        (root, "root"),
        (child1, "child1"),
        (grandchild1, "grandchild1"),
    ]
    grandchild2.get_hierarchy = lambda: [
        (root, "root"),
        (child2, "child2"),
        (grandchild2, "grandchild2"),
    ]

    return type(
        "NodeHierarchy",
        (),
        {
            "root": root,
            "child1": child1,
            "child2": child2,
            "grandchild1": grandchild1,
            "grandchild2": grandchild2,
        },
    )


def test_deepest_common_parent_single_node(node_hierarchy):
    result = Node.nearest_common_ancestor(node_hierarchy.child1)
    assert result == (node_hierarchy.child1, "child1")


def test_deepest_common_parent_common_parent(node_hierarchy):
    result = Node.nearest_common_ancestor(node_hierarchy.child1, node_hierarchy.child2)
    assert result == (node_hierarchy.root, "root")


def test_deepest_common_parent_different_depths(node_hierarchy):
    result = Node.nearest_common_ancestor(
        node_hierarchy.grandchild1, node_hierarchy.child2
    )
    assert result == (node_hierarchy.root, "root")


def test_deepest_common_parent_same_branch(node_hierarchy):
    result = Node.nearest_common_ancestor(
        node_hierarchy.grandchild1, node_hierarchy.child1
    )
    assert result == (node_hierarchy.child1, "child1")



from pydantic import BaseModel

from atopile.mcp.util import MCPTools
from faebryk.libs.backend.packages.api import PackagesAPIClient, _Endpoints

packages_tools = MCPTools()


class PackageInfoVeryBrief(BaseModel):
    identifier: str
    version: str
    summary: str


@packages_tools.register()
def inspect_package(
    identifier: str, version: str | None = None
) -> _Endpoints.PackageRelease.Response:
    """
    Get information for a published atopile community package,
    including readme if available.
    Details are sourced from the package's latest release.
    """
    client = PackagesAPIClient()
    package = client.get_package(identifier, version=version)

    return package


@packages_tools.register()
def find_packages(query: str) -> list[PackageInfoVeryBrief]:
    """
    Search for atopile community packages by identifier or summary,
    with closest matches appearing first.
    """
    client = PackagesAPIClient()
    packages = client.query_packages(query)

    return [
        PackageInfoVeryBrief(
            identifier=p.identifier, version=p.version, summary=p.summary
        )
        for p in packages.packages
    ]



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F
from faebryk.core.moduleinterface import ModuleInterface
from faebryk.libs.library import L
from faebryk.libs.units import P


class USB2_0(ModuleInterface):
    usb_if: F.USB2_0_IF

    def __preinit__(self):
        self.usb_if.buspower.voltage.constrain_subset(
            L.Range.from_center(5 * P.V, 0.25 * P.V)
        )



import fnmatch
import logging
import os
import re
import sys
from abc import ABC, abstractmethod
from contextlib import _GeneratorContextManager, contextmanager
from contextvars import ContextVar
from enum import Enum
from pathlib import Path
from typing import (
    Annotated,
    Any,
    Callable,
    Generator,
    Iterable,
    Literal,
    Self,
    Union,
    override,
)

from more_itertools import first
from pydantic import (
    AliasChoices,
    BaseModel,
    ConfigDict,
    Field,
    TypeAdapter,
    ValidationError,
    ValidationInfo,
    field_serializer,
    field_validator,
    model_validator,
)
from pydantic.networks import HttpUrl
from pydantic_settings import (
    BaseSettings,
    PydanticBaseSettingsSource,
    SettingsConfigDict,
    SettingsError,
    YamlConfigSettingsSource,
)
from ruamel.yaml import YAML

from atopile import address, version
from atopile.address import AddressError, AddrStr
from atopile.errors import (
    UserBadParameterError,
    UserException,
    UserFileNotFoundError,
    UserNoProjectException,
)
from atopile.version import (
    DISTRIBUTION_NAME,
    clean_version,
    get_installed_atopile_version,
)
from faebryk.libs.exceptions import UserResourceException
from faebryk.libs.test.testutil import in_test
from faebryk.libs.util import indented_container, md_list

logger = logging.getLogger(__name__)
yaml = YAML()


APPLICATION_NAME = "atopile"
ENV_VAR_PREFIX = "ATO_"
PROJECT_CONFIG_FILENAME = "ato.yaml"
GLOBAL_CONFIG_FILENAME = "config.yaml"


def _loc_to_dot_sep(loc: tuple[str | int, ...]) -> str:
    # via https://docs.pydantic.dev/dev/errors/errors/#customize-error-messages
    path = ""
    for i, x in enumerate(loc):
        if isinstance(x, str):
            if i > 0:
                path += "."
            path += x
        elif isinstance(x, int):
            path += f"[{x}]"
        else:
            raise TypeError("Unexpected type")
    return path


def _convert_errors(e: ValidationError) -> list[dict[str, Any]]:
    return [
        {"msg": error["msg"], "loc": _loc_to_dot_sep(error["loc"])}
        for error in e.errors()
    ]


def _try_construct_config[T](
    model: type[T], identifier: str | Path | None = None, **kwargs: Any
) -> T:
    message_prefix = f"`{identifier}`: " if identifier else ""

    try:
        return model(**kwargs)
    except ValidationError as ex:
        excs = ExceptionGroup(
            "Configuration is invalid",
            [
                UserConfigurationError(
                    f"{message_prefix}{error['msg']}: `{error['loc']}`"
                )
                for error in _convert_errors(ex)
            ],
        )
        raise excs from ex
    except SettingsError as ex:
        raise UserConfigurationError(f"Invalid config: {ex}") from ex


class BaseConfigModel(BaseModel):
    model_config = ConfigDict(use_attribute_docstrings=True)


class BuildType(Enum):
    ATO = "ato"
    PYTHON = "python"


class UserConfigurationError(UserException):
    """An error in the config file."""


class UserConfigNotFoundError(UserException):
    """No project config file was found."""


class ConfigFileSettingsSource(YamlConfigSettingsSource, ABC):
    def __init__(self, settings_cls: type[BaseSettings]):
        self.yaml_file_path = self.find_config_file()
        self.yaml_file_encoding = "utf-8"
        self.yaml_data = self._read_files(self.yaml_file_path)

        super(YamlConfigSettingsSource, self).__init__(settings_cls, self.get_data())

    @classmethod
    @abstractmethod
    def find_config_file(cls) -> Path | None: ...

    @abstractmethod
    def get_data(self) -> dict[str, Any]: ...


class GlobalConfigSettingsSource(ConfigFileSettingsSource):
    @classmethod
    def find_config_file(cls) -> Path | None:
        """Find the global config file in the user's home directory."""

        # note deliberate use of ~/.config on all platforms
        # (rather than e.g. platformdirs)
        config_dir = (
            Path(os.environ.get("XDG_CONFIG_HOME", Path.home() / ".config"))
            / APPLICATION_NAME
        )
        config_file = config_dir / GLOBAL_CONFIG_FILENAME
        return config_file if config_file.exists() else None

    def get_data(self) -> dict[str, Any]:
        return self.yaml_data if self.yaml_data else {}


def _find_project_dir(start: Path) -> Path | None:
    """
    Search parent directories, up to the root, for a directory containing a project
    config file.
    """
    path = start
    while not (path / PROJECT_CONFIG_FILENAME).is_file():
        path = path.parent
        if path == path.parent:
            return None

    return path.resolve().absolute()


def _find_project_config_file(start: Path) -> Path | None:
    """Search parent directories, up to the root, for a project config file."""
    if (project_dir := _find_project_dir(start)) is not None:
        return project_dir / PROJECT_CONFIG_FILENAME

    return None


class ProjectConfigSettingsSource(ConfigFileSettingsSource):
    @classmethod
    def find_config_file(cls) -> Path | None:
        """
        Find a project config file in the specified directory or any parent directories.
        """
        if _project_dir:
            return _project_dir / PROJECT_CONFIG_FILENAME

        return _find_project_config_file(Path.cwd())

    def get_data(self) -> dict[str, Any]:
        return self.yaml_data or {}


class ProjectPaths(BaseConfigModel):
    """
    Project global paths
    """

    root: Path
    """Project root directory (where the ato.yaml file is located)"""

    src: Path
    """Project source code directory"""

    parts: Path
    """Source code directory for parts"""

    layout: Path
    """Project layout directory where KiCAD projects are stored and searched for"""

    build: Path
    """Build artifact output directory"""

    logs: Path
    """Build logs directory"""

    manifest: Path
    """Build manifest file"""

    modules: Path
    """Project modules directory (`.ato/modules` from the project root)"""

    # TODO: remove, deprecated
    footprints: Path | None = None
    """Deprecated: Project footprints directory"""

    # TODO: remove, deprecated
    component_lib: Path | None = None
    """Deprecated: Component library directory for builds"""

    def __init__(self, **data: Any):
        data.setdefault("root", _project_dir or Path.cwd())
        data["src"] = Path(data.get("src", data["root"] / "elec" / "src"))
        data.setdefault("parts", data["src"] / "parts")
        data.setdefault("layout", data["root"] / "elec" / "layout")
        data["build"] = Path(data.get("build", data["root"] / "build"))
        data.setdefault("logs", data["build"] / "logs")
        data.setdefault("manifest", data["build"] / "manifest.json")
        data.setdefault("modules", data["root"] / ".ato" / "modules")

        super().__init__(**data)

    @model_validator(mode="after")
    def make_paths_absolute(model: "ProjectPaths") -> "ProjectPaths":
        """Make all paths absolute relative to the project root."""
        if not model.root.is_absolute():
            model.root = model.root.resolve().absolute()

        for field_name, field_value in model:
            if field_name != "root" and isinstance(field_value, Path):
                if not field_value.is_absolute():
                    setattr(model, field_name, model.root / field_value)
                setattr(
                    model, field_name, getattr(model, field_name).resolve().absolute()
                )
        return model

    @model_validator(mode="before")
    def check_deprecated_fields(cls, data: dict[str, Any]) -> dict[str, Any]:
        if "component_lib" in data and data["component_lib"] is not None:
            raise UserConfigurationError(
                "The 'component_lib' field in your ato.yaml is deprecated. "
                "Delete your component_lib and the entry in the ato.yaml. "
                "Use 'parts' instead."
            )
        if "footprints" in data and data["footprints"] is not None:
            raise UserConfigurationError(
                "The 'footprints' field in your ato.yaml is deprecated. "
                "If you are manually assigning hand-made footprints you "
                "probably want to make atomic parts instead. "
                "Please remove it."
            )
        return data

    def ensure(self) -> None:
        self.build.mkdir(parents=True, exist_ok=True)
        self.layout.mkdir(parents=True, exist_ok=True)


class BuildTargetPaths(BaseConfigModel):
    """
    Build-target specific paths
    """

    layout: Path
    """Build-target layout file"""

    output_base: Path
    """Extension-less filename for build artifacts"""

    netlist: Path
    """Build-target netlist file"""

    fp_lib_table: Path
    """Project footprint library table file"""

    kicad_project: Path
    """Build-target KiCAD project file"""

    def __init__(self, name: str, project_paths: ProjectPaths, **data: Any):
        if layout_data := data.get("layout"):
            data["layout"] = BuildTargetPaths.find_layout(Path(layout_data))
        else:
            data["layout"] = BuildTargetPaths.find_layout(
                project_paths.root / project_paths.layout / name
            )

        if output_base_data := data.get("output_base"):
            data["output_base"] = Path(output_base_data)
        else:
            data["output_base"] = project_paths.build / "builds" / name / name
            data["output_base"].parent.mkdir(parents=True, exist_ok=True)

        data.setdefault("netlist", data["output_base"] / f"{name}.net")
        data.setdefault("fp_lib_table", data["layout"].parent / "fp-lib-table")
        data.setdefault("kicad_project", data["layout"].with_suffix(".kicad_pro"))
        # We deliberately don't set a root for an individual build
        super().__init__(**data)

    @classmethod
    def find_layout(cls, layout_base: Path) -> Path:
        """Find the layout associated with a build."""

        if layout_base.with_suffix(".kicad_pcb").exists():
            return layout_base.with_suffix(".kicad_pcb").resolve().absolute()
        elif layout_base.is_dir():
            layout_candidates = list(
                filter(
                    BuildTargetPaths.match_user_layout, layout_base.glob("*.kicad_pcb")
                )
            )

            if len(layout_candidates) == 1:
                return layout_candidates[0].resolve().absolute()
            elif len(layout_candidates) > 1:
                raise UserException(
                    "Layout directories must contain only 1 layout,"
                    f" but {len(layout_candidates)} found in {layout_base}"
                )

        # default location, to create later
        return layout_base.resolve().absolute() / f"{layout_base.name}.kicad_pcb"

    def ensure_layout(self):
        """Return the layout associated with a build."""
        if not self.layout.exists():
            logger.info("Creating new layout at %s", self.layout)
            self.layout.parent.mkdir(parents=True, exist_ok=True)

            # delayed import to improve startup time
            from faebryk.libs.kicad.fileformats_latest import C_kicad_pcb_file

            C_kicad_pcb_file.skeleton(
                generator=DISTRIBUTION_NAME,
                generator_version=str(get_installed_atopile_version()),
            ).dumps(self.layout)
        elif not self.layout.is_file():
            raise UserResourceException(f"Layout is not a file: {self.layout}")

    @classmethod
    def match_user_layout(cls, path: Path) -> bool:
        """Check whether a given filename is a KiCAD autosaved layout."""
        autosave_patterns = [
            "_autosave-*",
            "*-save.kicad_pcb",
        ]

        name = path.name

        for pattern in autosave_patterns:
            if fnmatch.fnmatch(name, pattern):
                return False
        return True


class BuildTargetConfig(BaseConfigModel, validate_assignment=True):
    _project_paths: ProjectPaths

    name: str
    address: str | None = Field(alias="entry")
    """
    Entry point or root node for the build-target.
    Everything that exists within this module will be built as part of this build-target
    """

    targets: list[str] = Field(default=["__default__"])  # TODO: validate
    """A list of targets' names to build after updating the layout"""

    exclude_targets: list[str] = Field(default=[])
    """
    A list of targets' names to exclude.

    Applied after `targets` are specified. This makes it useful to exclude
    targets that aren't relevant to a specific build - especially in CI,
    which typically builds **all** targets.
    """

    exclude_checks: list[str] = Field(default=[])
    """
    A list of checks to exclude.

    Use qualified name of check e.g
        - `PCB.requires_drc_check`
        - `I2C.requires_unique_addresses`
        - `requires_external_usage`
    """
    fail_on_drcs: bool = Field(default=False)
    dont_solve_equations: bool = Field(default=False)
    keep_designators: bool | None = Field(default=True)
    keep_picked_parts: bool | None = Field(default=None)
    keep_net_names: bool | None = Field(default=None)
    frozen: bool = Field(default=False)
    hide_designators: bool | None = Field(default=False)
    paths: BuildTargetPaths

    def __init__(self, **data: Any):
        super().__init__(**data)
        self._project_paths = data.get("_project_paths", ProjectPaths())

    @model_validator(mode="before")
    def init_paths(cls, data: dict) -> dict:
        match data.get("paths"):
            case dict() | None:
                data["paths"] = BuildTargetPaths(
                    name=data["name"],
                    project_paths=data["_project_paths"],
                    **data.get("paths", {}),
                )
            case BuildTargetPaths():
                pass
            case _:
                raise ValueError(f"Invalid build paths: {data.get('paths')}")
        return data

    def _set_frozen(self, frozen: bool):
        if self.keep_designators is None:
            self.keep_designators = frozen
        if self.keep_picked_parts is None:
            self.keep_picked_parts = frozen
        if self.keep_net_names is None:
            self.keep_net_names = frozen
        self.frozen = frozen

    @model_validator(mode="after")
    def validate_frozen_after(self) -> Self:
        if self.frozen:
            if not self.keep_designators:
                raise ValueError(
                    "`keep_designators` must be true when `frozen` is true"
                )
            if not self.keep_picked_parts:
                raise ValueError(
                    "`keep_picked_parts` must be true when `frozen` is true"
                )
            if not self.keep_net_names:
                raise ValueError("`keep_net_names` must be true when `frozen` is true")

        return self

    @property
    def build_type(self) -> BuildType:
        """
        Determine build type from the entry.

        **/*.ato:* -> BuildType.ATO
        **/*.py:* -> BuildType.PYTHON
        """
        suffix = self.entry_file_path.suffix

        match suffix:
            case ".ato":
                return BuildType.ATO
            case ".py":
                return BuildType.PYTHON
            case _:
                raise UserException(
                    f"Unknown entry suffix: {suffix} for {self.entry_section}"
                )

    @property
    def entry_file_path(self) -> Path:
        """An absolute path to the entry file."""
        address = AddrStr(self.address)
        return self._project_paths.root / address.file_path

    @property
    def entry_section(self) -> str:
        """The path to the entry module."""
        address = AddrStr(self.address)
        return address.entry_section

    def ensure(self):
        """Ensure this build config is ready to be used"""
        self.paths.ensure_layout()


class DependencySpec(BaseConfigModel):
    type: str
    # TODO ugly af, because we are mixing specs and config
    # should be config only, not in spec
    identifier: str

    @staticmethod
    def from_str(spec_str: str) -> "DependencySpec":
        if "://" not in spec_str:
            spec_str = "registry://" + spec_str
            # TODO default registry

        type_specifier, _ = spec_str.split("://", 1)

        # TODO dont use hardcoded strings
        if type_specifier.startswith("file"):
            return FileDependencySpec.from_str(spec_str)
        elif type_specifier.startswith("git"):
            return GitDependencySpec.from_str(spec_str)
        elif type_specifier.startswith("registry"):
            return RegistryDependencySpec.from_str(spec_str)
        else:
            raise UserConfigurationError(
                f"Invalid type specifier: {type_specifier} in {spec_str}"
            )

    def matches(self, other: "DependencySpec") -> bool:
        return self.identifier == other.identifier


class FileDependencySpec(DependencySpec):
    type: Literal["file"] = "file"
    path: Path
    identifier: str | None = None

    @field_serializer("path")
    def serialize_path(self, path: Path, _info: Any) -> str:
        return str(path)

    @staticmethod
    @override
    def from_str(spec_str: str) -> "FileDependencySpec":
        _, path = spec_str.split("://", 1)
        return FileDependencySpec(path=Path(path))


class GitDependencySpec(DependencySpec):
    type: Literal["git"] = "git"
    repo_url: str
    path_within_repo: Path | None = None
    ref: str | None = None
    identifier: str | None = None

    @staticmethod
    @override
    def from_str(spec_str: str) -> "GitDependencySpec":
        # Pattern to match git dependency spec format: git://<repo_url>.git[#<ref>][:<path_within_repo>]
        # - repo_url: everything after git:// until # or :
        # - ref: optional, everything between # and : (if present)
        # - path_within_repo: optional, everything after :
        pattern = (
            r"^git://"  # Protocol prefix
            r"(?P<repo_url>.+?\.git)"  # Repository URL (non-greedy match until .git)
            r"(?:#(?!:|$)"  # Optional ref part: '#' not followed by ':' or end
            r"(?P<ref>[^:]+)"  # Reference value (anything not a colon)
            r")?"  # End of optional ref group
            r"(?::(?P<path_within_repo>.*))?"  # Optional path part: ':' followed by
            # the path (colon not captured)
            r"$"  # End of string
        )
        match = re.match(pattern, spec_str)
        if not match:
            raise ValueError(f"Invalid git dependency spec: {spec_str}")

        repo_url = match.group("repo_url")
        ref = match.group("ref")
        path_within_repo = match.group("path_within_repo")
        if path_within_repo is not None:
            path_within_repo = Path(path_within_repo)

        return GitDependencySpec(
            repo_url=repo_url,
            ref=ref,
            path_within_repo=path_within_repo,
        )


class RegistryDependencySpec(DependencySpec):
    type: Literal["registry"] = "registry"
    release: str | None = None

    @property
    @override
    def identifier(self) -> str:
        return self.identifier

    @staticmethod
    @override
    def from_str(spec_str: str) -> "RegistryDependencySpec":
        _, identifier = spec_str.split("://", 1)
        if "@" in identifier:
            identifier, release = identifier.split("@", 1)
        else:
            release = None
        return RegistryDependencySpec(identifier=identifier, release=release)


_DependencySpec = Annotated[
    Union[FileDependencySpec, GitDependencySpec, RegistryDependencySpec],
    Field(discriminator="type"),
]


class ServicesConfig(BaseConfigModel):
    class Components(BaseConfigModel):
        url: str = Field(default="https://components.atopileapi.com")
        """Components URL"""

    class Packages(BaseConfigModel):
        url: str = Field(default="https://packages.atopileapi.com")
        """Packages URL"""

    @field_validator("components", mode="before")
    def validate_components(cls, value: Components | str) -> Components:
        # also accepts a string for backwards compatibility
        if isinstance(value, str):
            return cls.Components(url=value)
        return value

    components: Components = Field(default_factory=Components)
    packages: Packages = Field(default_factory=Packages)


# TODO: expand
RequirementSpec = Annotated[
    str,
    Field(
        pattern=r"^((>|>=|<|<=|\^)?[0-9]+\.[0-9]+\.[0-9]+|(>|>=)[0-9]+\.[0-9]+\.[0-9]+,(<|<=)[0-9]+\.[0-9]+\.[0-9]+)$"
    ),
]


class PackageConfig(BaseConfigModel):
    """Defines a package"""

    class Author(BaseConfigModel):
        name: str
        email: str

    identifier: str = Field(
        pattern=r"^(?P<owner>[a-zA-Z0-9](?:[a-zA-Z0-9]|(-[a-zA-Z0-9])){0,38})/(?P<name>[a-z][a-z0-9\-]+)(/(?P<subpackage>[a-z][a-z0-9\-]+))?$"
    )
    """
    The qualified name of the project, as it'd be installed from a package manager.
    eg. `pepper/my-project` or `pepper/my-project/sub-package`
    May contain numbers and lowercase ASCII letters only. The owner must match the
    GitHub organization from which the project is published (eg. `pepper`).
    Required for publishing.
    """

    version: str = Field(
        # semver subset only
        pattern=r"^(?P<major>[0-9]+)\.(?P<minor>[0-9]+)\.(?P<patch>[0-9]+)$",
        default="0.0.0",
    )
    """
    Project version, formatted according to the SemVer specification. See https://semver.org/.
    Contains MAJOR.MINOR.PATCH only, e.g. 1.0.1
    Required for publishing.
    """

    repository: HttpUrl | None = Field(default=None)
    """
    The repository URL of the project.
    Required for publishing.
    """

    authors: list[Author] | None = Field(default=None)
    """
    List of project authors.
    Required for publishing.
    """

    license: str | None = Field(default=None)
    """
    The project's license, as an SPDX 2.3 license expression.
    Required for publishing.
    """

    summary: str | None = Field(default=None)
    """
    A short blurb about the project.
    Required for publishing.
    """

    homepage: str | None = Field(default=None)
    """
    The project's homepage, if separate from the repository.
    """

    readme: str | None = Field(default="README.md")
    """
    Path to the project's README file, relative to this `ato.yaml`
    """


class ProjectConfig(BaseConfigModel):
    """Project-level config"""

    requires_atopile: RequirementSpec = Field(
        validation_alias=AliasChoices("requires-atopile", "requires_atopile"),
        serialization_alias="requires-atopile",
        default=f"^{clean_version(version.get_installed_atopile_version())}",
    )
    """
    Version required to build this project.
    """

    @model_validator(mode="before")
    def check_deprecated_fields(cls, values: dict[str, Any]) -> dict[str, Any]:
        if "ato-version" in values:
            raise UserConfigurationError(
                "The 'ato-version' field in your ato.yaml is deprecated. "
                "Use 'requires-atopile' instead."
            )
        return values

    package: PackageConfig | None = Field(default=None)
    """
    Defines a package
    """

    paths: ProjectPaths = Field(default_factory=ProjectPaths)
    dependencies: list[_DependencySpec] | None = Field(default=None)
    """
    Represents requirements on other projects.

    Typically, you shouldn't modify this directly.

    Instead, use the `ato add/remove <package>` commands.
    """

    entry: str | None = Field(default=None)
    builds: dict[str, BuildTargetConfig] = Field(default_factory=dict)
    """A map of all the build targets (/ "builds") in this project."""

    services: ServicesConfig = Field(default_factory=ServicesConfig)
    open_layout_on_build: bool = Field(default=False)
    """Automatically open pcbnew when applying netlist"""

    @classmethod
    def from_path(cls, path: Path | None) -> "ProjectConfig | None":
        if path is None:
            return None

        config_file = path / PROJECT_CONFIG_FILENAME

        if not config_file.exists():
            return None

        try:
            file_contents = yaml.load(config_file)
        except FileNotFoundError as e:
            raise UserFileNotFoundError(f"Failed to load project config: {e}") from e
        except Exception as e:
            raise UserConfigurationError(f"Failed to load project config: {e}") from e

        file_contents.setdefault("paths", {}).setdefault("root", path)
        return _try_construct_config(
            ProjectConfig, identifier=config_file, **file_contents
        )

    @classmethod
    def from_bytes(cls, data: bytes) -> "ProjectConfig":
        try:
            file_contents = yaml.load(data)
        except Exception as e:
            raise UserConfigurationError(f"Failed to load project config: {e}") from e

        return _try_construct_config(ProjectConfig, identifier="", **file_contents)

    @field_validator("builds", mode="before")
    def init_builds(
        cls, value: dict[str, dict[str, Any] | BuildTargetConfig], info: ValidationInfo
    ) -> dict[str, Any]:
        for build_name, data in value.items():
            match data:
                case BuildTargetConfig():
                    data.name = build_name
                case dict():
                    data.setdefault("name", build_name)
                    data["_project_paths"] = info.data["paths"]
                case _:
                    raise ValueError(f"Invalid build data: {data}")
        return value

    @field_validator("dependencies", mode="before")
    def validate_dependencies(
        cls, value: list[dict[str, Any]] | None
    ) -> list[DependencySpec]:
        return [
            DependencySpec.from_str(dep)
            if isinstance(dep, str)
            else TypeAdapter(_DependencySpec).validate_python(dep)
            for dep in (value or [])
        ]

    @classmethod
    def skeleton(cls, entry: str, paths: ProjectPaths | None):
        """Creates a minimal ProjectConfig"""
        project_paths = paths or ProjectPaths()
        return _try_construct_config(
            ProjectConfig,
            paths=project_paths,
            entry=entry,
            builds={
                "default": BuildTargetConfig(
                    _project_paths=project_paths,
                    name="default",
                    entry="",
                    paths=BuildTargetPaths(name="default", project_paths=project_paths),
                )
            },
        )

    @staticmethod
    def set_or_add_dependency(config: "Config", dependency: DependencySpec):
        def _add_dependency(config_data, _):
            # validate_dependencies is the validator that loads the dependencies
            # from the config file. It ensures the format of the ato.yaml
            deps = ProjectConfig.validate_dependencies(
                config_data.get("dependencies", [])
            )  # type: ignore (class method)

            serialized = dependency.model_dump(mode="json")

            for i, dep in enumerate(deps):
                if dep.matches(dependency):
                    config_data["dependencies"][i] = serialized
                    break
            else:
                if config_data.get("dependencies") is None:
                    config_data["dependencies"] = []
                config_data["dependencies"].append(serialized)

            return config_data

        config.update_project_settings(_add_dependency, {})

    @staticmethod
    def remove_dependency(config: "Config", dependency: DependencySpec):
        def _remove_dependency(config_data, _):
            deps = ProjectConfig.validate_dependencies(
                config_data.get("dependencies", [])
            )  # type: ignore (class method)

            for i, dep in enumerate(deps):
                if dep.matches(dependency):
                    del config_data["dependencies"][i]
                    break

            return config_data

        config.update_project_settings(_remove_dependency, {})

    def get_relative_to_kicad_project(self, path: Path) -> Path:
        if not self.builds:
            if in_test():
                raise ValueError(
                    "No builds found. Did you forget: "
                    "`@pytest.mark.usefixtures('setup_project_config')`?"
                )
            raise UserConfigurationError("No builds found in project config")

        rel_paths = {
            name: path.relative_to(build.paths.kicad_project.parent, walk_up=True)
            for name, build in self.builds.items()
        }
        # TODO check this
        if len(set(rel_paths.values())) != 1:
            raise ValueError(
                "All builds must have the same common prefix path: "
                f"{indented_container(rel_paths)}"
            )
        return Path("${KIPRJMOD}") / first(rel_paths.values())


class ProjectSettings(ProjectConfig, BaseSettings):  # FIXME
    """
    Project-level config, loaded from
    - global config e.g. ~/.config/atopile/config.yaml
    - ato.yaml
    - environment variables e.g. ATO_SERVICES_COMPONENTS_API_URL
    """

    # TOOD: ignore but warn for extra fields
    model_config = BaseConfigModel.model_config | SettingsConfigDict(
        env_prefix=ENV_VAR_PREFIX,
        env_nested_delimiter="_",
        enable_decoding=False,
        use_attribute_docstrings=True,
        extra="forbid",
    )

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> tuple[PydanticBaseSettingsSource, ...]:
        return (
            init_settings,
            env_settings,
            ProjectConfigSettingsSource(settings_cls),
            GlobalConfigSettingsSource(settings_cls),
        )


_current_build_cfg: ContextVar[BuildTargetConfig | None] = ContextVar(
    "current_build_cfg", default=None
)


@contextmanager
def _build_context(config: "Config", build_name: str):
    cfg = config.project.builds[build_name]
    cfg.ensure()
    token = _current_build_cfg.set(cfg)
    try:
        yield
    finally:
        _current_build_cfg.reset(token)


class Config:
    _project: ProjectSettings | ProjectConfig | None
    _entry: str | None
    _selected_builds: list[str] | None
    _project_dir: Path | None
    interactive: bool

    def __init__(self) -> None:
        self._project_dir = _project_dir
        self._project = _try_construct_config(ProjectSettings)
        self._entry = None
        self._selected_builds = None
        # Check if we're in an interactive terminal session (cross-platform)
        try:
            self.interactive = sys.stdout.isatty() and sys.stdin.isatty()
        except (AttributeError, ValueError):
            # If we can't determine, default to True for better user experience
            self.interactive = True

    def __repr__(self) -> str:
        return self._project.__repr__()

    def __rich_repr__(self):
        yield "project", self._project
        yield "entry", self._entry
        yield "selected_builds", self._selected_builds
        yield "project_dir", self._project_dir

    @property
    def project(self) -> ProjectSettings | ProjectConfig:
        if self._project is None:
            # TODO: better message
            raise UserConfigurationError("No project config found")

        return self._project

    @project.setter
    def project(self, value: ProjectSettings | ProjectConfig) -> None:
        self._project = value

    @property
    def project_dir(self) -> Path:
        if not self._project_dir:
            raise ValueError("No project directory found")
        return self._project_dir

    @project_dir.setter
    def project_dir(self, value: os.PathLike) -> None:
        global _project_dir
        _project_dir = Path(value)

        if not (_project_dir / PROJECT_CONFIG_FILENAME).is_file():
            raise UserConfigNotFoundError(
                f"No `{PROJECT_CONFIG_FILENAME}` found in the specified directory"
            )

        self._project_dir = _project_dir
        self._project = _try_construct_config(ProjectSettings)

    def update_project_settings(
        self, transformer: Callable[[dict, dict], dict], new_data: dict
    ) -> None:
        """Apply an update to the project config file."""
        yaml = YAML(typ="rt")  # round-trip
        filename = self.project_dir / PROJECT_CONFIG_FILENAME
        temp_filename = filename.with_suffix(".yaml.tmp")

        try:
            with filename.open("r", encoding="utf-8") as file:
                yaml_data: dict = yaml.load(filename) or {}

            yaml_data = transformer(yaml_data, new_data)

            with temp_filename.open("w", encoding="utf-8") as file:
                yaml.dump(yaml_data, file)

            temp_filename.replace(filename)

            self._project = _try_construct_config(ProjectSettings)

        except Exception as e:
            try:
                temp_filename.unlink(missing_ok=True)
            except Exception:
                pass
            raise e

    @property
    def selected_builds(self) -> Iterable[str]:
        return self._selected_builds or self.project.builds.keys()

    @selected_builds.setter
    def selected_builds(self, value: list[str]) -> None:
        self._selected_builds = value

    @property
    def builds(self) -> Generator[_GeneratorContextManager[None], None, None]:
        """Return an iterable of BuildContext objects for each build."""
        return (_build_context(self, name) for name in self.selected_builds)

    def select_build(self, name: str) -> _GeneratorContextManager[None]:
        return _build_context(self, name)

    @property
    def build(self) -> BuildTargetConfig:
        if current := self._current_build:
            return current
        raise RuntimeError("No build config is currently active")

    @property
    def has_project(self) -> bool:
        try:
            project_dir = self.project_dir
        except ValueError:
            return False
        return (project_dir / PROJECT_CONFIG_FILENAME).exists()

    @property
    def _current_build(self) -> BuildTargetConfig | None:
        return _current_build_cfg.get()

    def _setup_standalone(self, entry: str | None, entry_arg_file_path: Path) -> None:
        if not entry:
            raise UserBadParameterError(
                "You must specify an entry to build with the --standalone option"
            )
        if not entry_arg_file_path.exists():
            raise UserBadParameterError(
                f"The file you have specified does not exist: {entry_arg_file_path}"
            )

        if not entry_arg_file_path.is_file():
            raise UserBadParameterError(
                "The path you're building with the --standalone"
                f" option must be a file {entry_arg_file_path}",
                markdown=False,
            )

        if config.has_project:
            raise UserBadParameterError(
                "Project config must not be present for standalone builds"
            )

        try:
            AddrStr(entry).entry_section
        except AddressError:
            raise UserBadParameterError(
                "You must specify what to build within a file to build with the"
                " --standalone option"
            )

        # don't trigger reload
        self._project_dir = entry_arg_file_path.parent
        root = self._project_dir
        standalone_dir = root / f"standalone_{entry_arg_file_path.stem}"
        self._project = ProjectConfig.skeleton(
            entry=entry,
            paths=ProjectPaths(
                root=root,
                src=root,
                layout=standalone_dir / "layout",
                build=standalone_dir / "build",
                # TODO definitely need option to override this
                parts=standalone_dir / "parts",
            ),
        )

    def _check_entry_arg_file_path(
        self, entry: str | None, entry_arg_file_path: Path
    ) -> AddrStr | None:
        entry_addr_override = None

        if entry:
            if entry_arg_file_path.is_file():
                if entry_section := address.get_entry_section(AddrStr(entry)):
                    entry_addr_override = address.from_parts(
                        str(entry_arg_file_path.absolute()),
                        entry_section,
                    )
                else:
                    raise UserBadParameterError(
                        "If an entry of a file is specified, you must specify"
                        " the node within it you want to build.",
                        title="Bad 'entry' parameter",
                    )

            elif entry_arg_file_path.is_dir():
                pass

            elif not entry_arg_file_path.exists():
                raise UserBadParameterError(
                    "The entry you have specified does not exist.",
                    title="Bad 'entry' parameter",
                )
            else:
                raise ValueError(
                    f"Unexpected entry path type {entry_arg_file_path}"
                    " - this should never happen!"
                )

        return entry_addr_override

    def _get_entry_arg_file_path(
        self, entry: str | None, working_dir: Path | None
    ) -> tuple[AddrStr | None, Path]:
        # basic the entry address if provided, otherwise leave it as None

        if entry is None:
            entry_arg_file_path = working_dir or Path.cwd()
        else:
            entry = AddrStr(entry)

            if address.get_file(entry) is None:
                raise UserBadParameterError(
                    f"Invalid entry address {entry} - entry must specify a file.",
                    title="Bad 'entry' parameter",
                )

            entry_arg_file_path = (
                Path(address.get_file(entry)).expanduser().resolve().absolute()
            )

        return entry, entry_arg_file_path

    def apply_options(
        self,
        entry: str | None,
        standalone: bool = False,
        target: Iterable[str] = (),
        selected_builds: Iterable[str] = (),
        frozen: bool | None = None,
        working_dir: Path | None = None,
        **kwargs: Any,
    ) -> None:
        if working_dir:
            working_dir = Path(working_dir).expanduser().resolve().absolute()

        entry, entry_arg_file_path = self._get_entry_arg_file_path(entry, working_dir)

        if standalone:
            self._setup_standalone(entry, entry_arg_file_path)
        else:
            if config_file_path := _find_project_config_file(entry_arg_file_path):
                self.project_dir = config_file_path.parent
            elif entry is None:
                raise UserNoProjectException(search_path=entry_arg_file_path)

            else:
                raise UserBadParameterError(
                    f"Specified entry path is not a file or directory: "
                    f"{entry_arg_file_path}",
                    markdown=False,
                )

        self.project.entry = entry

        logger.info("Using project %s", self.project_dir)

        # if we set an entry-point, we now need to deal with that
        entry_addr_override = self._check_entry_arg_file_path(
            entry, entry_arg_file_path
        )

        if self.project.paths is not None:
            self.project.paths.ensure()

        if selected_builds:
            self.selected_builds = list(selected_builds)

        for build_name in self.selected_builds:
            try:
                build_cfg = self.project.builds[build_name]
            except KeyError:
                raise UserBadParameterError(
                    f"Build `{build_name}` not found in project config.\n\nAvailable"
                    " builds:\n" + md_list(self.project.builds.keys())
                )

            if entry_addr_override is not None:
                build_cfg.address = entry_addr_override
            if target:
                build_cfg.targets = list(target)

            # Attach CLI options passed via kwargs
            for key, value in kwargs.items():
                if value is not None:
                    setattr(build_cfg, key, value)

            if frozen is not None:
                try:
                    build_cfg._set_frozen(frozen)
                except ValidationError as e:
                    # TODO: better error message
                    raise UserBadParameterError(
                        f"Invalid value for `frozen`: {e}",
                        title="Bad 'frozen' parameter",
                    )

    def should_open_layout_on_build(self) -> bool:
        """Returns whether atopile should open the layout after building"""
        # If the project config has an explicit setting, use that
        if self.project is not None and self.project.open_layout_on_build is not None:
            return self.project.open_layout_on_build

        # Otherwise, default to opening the layout if we're only building a
        # single target and we're running interactively
        return (
            (self.project is None or self.project.open_layout_on_build is None)
            and len(list(self.selected_builds)) == 1
            and self.interactive
        )


_project_dir: Path | None = None
config: Config = Config()



import logging
import shutil
import subprocess
from pathlib import Path

from faebryk.libs.util import (
    global_lock,
    has_uncommitted_changes,
    is_editable_install,
    run_live,
)

logger = logging.getLogger(__name__)


def check_for_recompile():
    if not is_editable_install():
        return

    THIS_DIR = Path(__file__).parent
    parser_files = THIS_DIR.rglob("Ato*")
    uncommitted_changes = has_uncommitted_changes(parser_files)
    # if can't check git, we need to assume changes in editable mode
    if uncommitted_changes is not None and not uncommitted_changes:
        return

    # binary provided by antlr4-tools python package
    if shutil.which("antlr4") is None:
        logger.warning("antlr-tools not found, did you run `uv sync --dev`?")
        return

    logger.warning("Recompiling ato grammar")

    # compile using antlr4 python package
    GRAMMAR_FILES = [
        "AtoLexer.g4",
        "AtoParser.g4",
    ]

    with global_lock(THIS_DIR / "antlr.lock", timeout_s=10):
        try:
            run_live(
                [
                    "antlr4",
                    "-long-messages",
                    "-visitor",
                    "-no-listener",
                    "-Dlanguage=Python3",
                    *GRAMMAR_FILES,
                ],
                cwd=THIS_DIR,
            )
        except subprocess.CalledProcessError as e:
            logger.error(f"ANTLR compilation failed:\n{e.stderr}")


check_for_recompile()



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

from collections.abc import Iterable, Iterator
from enum import Enum
from typing import TYPE_CHECKING, Any, Protocol, overload, override, runtime_checkable

from faebryk.libs.units import Unit, dimensionless
from faebryk.libs.util import (
    Serializable,
    SerializableEnum,
)

if TYPE_CHECKING:
    from faebryk.libs.sets.quantity_sets import (
        Quantity_Interval_Disjoint,
        QuantitySetLike,
    )


# Protocols ----------------------------------------------------------------------------
@runtime_checkable
class P_Set[T](Serializable, Protocol):
    def is_empty(self) -> bool: ...

    def is_finite(self) -> bool: ...

    def __bool__(self) -> bool:
        raise Exception("don't use bool to check for emptiness, use is_empty()")

    def __contains__(self, item: T) -> bool: ...

    @staticmethod
    def from_value(value) -> "P_Set":
        from faebryk.libs.sets.quantity_sets import (
            Quantity_Interval_Disjoint,
            QuantitySetLikeR,
        )

        if isinstance(value, QuantitySetLikeR) and not isinstance(value, bool):
            return Quantity_Interval_Disjoint.from_value(value)
        if isinstance(value, (bool, BoolSet)):
            return BoolSet.from_value(value)
        if (
            isinstance(value, Enum)
            or isinstance(value, type)
            and issubclass(value, Enum)
        ):
            return EnumSet(value)
        if isinstance(value, PlainSet):
            return value
        raise ValueError(f"cannot convert {value} to P_Set")

    def is_subset_of(self, other: "P_Set[T]") -> bool:
        assert type(self) is not P_Set
        return self.is_subset_of(other)

    def is_superset_of(self, other: "P_Set[T]") -> bool:
        return P_Set.from_value(other).is_subset_of(self)

    @staticmethod
    def intersect_all[P: P_Set](*sets: P) -> P:
        if not sets:
            raise ValueError("no sets to intersect")
            # return PlainSet()
        out = sets[0]
        for s in sets[1:]:
            out = out & s
        return out

    def __and__[P: P_Set](self: P, other: P) -> P: ...

    def is_single_element(self) -> bool: ...

    def any(self) -> T: ...

    def serialize(self) -> dict:
        return {"type": type(self).__name__, "data": self.serialize_pset()}

    def serialize_pset(self) -> dict: ...

    @classmethod
    def deserialize(cls, data: dict) -> "P_Set":
        from faebryk.libs.sets.numeric_sets import (
            Numeric_Interval,
            Numeric_Interval_Disjoint,
        )
        from faebryk.libs.sets.quantity_sets import (
            Quantity_Interval,
            Quantity_Interval_Disjoint,
            Quantity_Set_Discrete,
        )

        types = [
            Numeric_Interval,
            Numeric_Interval_Disjoint,
            BoolSet,
            EnumSet,
            Quantity_Interval,
            Quantity_Interval_Disjoint,
            Quantity_Set_Discrete,
        ]
        cls_ = next((t for t in types if t.__name__ == data["type"]), None)
        if cls_ is None:
            raise ValueError(f"unknown type {data['type']}")
        return cls_.deserialize_pset(data["data"])

    @classmethod
    def deserialize_pset(cls, data: dict) -> "P_Set": ...


class P_IterableSet[T, IterT](P_Set[T], Iterable[IterT], Protocol): ...


class P_UnitSet[T](P_Set[T], Protocol):
    units: Unit


class P_IterableUnitSet[T, IterT](P_UnitSet[T], Iterable[IterT], Protocol): ...


# TODO consider making abstract
class PlainSet[U](P_IterableUnitSet[U, U]):
    def __init__(self, *elements: "U | PlainSet[U]"):
        self.units = dimensionless

        self.elements: frozenset[U] = frozenset(  # type: ignore
            {v for v in elements if not isinstance(v, PlainSet)}
            | {v for vb in elements if isinstance(vb, PlainSet) for v in vb.elements}
        )

    def is_empty(self) -> bool:
        return len(self.elements) == 0

    def __hash__(self) -> int:
        return hash(self.elements)

    def __repr__(self) -> str:
        return f"{type(self).__name__}({', '.join(repr(e) for e in self.elements)})"

    def __str__(self) -> str:
        # TODO move enum stuff to EnumSet
        return f"[{
            ', '.join(
                str(e) if not isinstance(e, Enum) else f'{e.name}'
                for e in self.elements
            )
        }]"

    def __iter__(self) -> Iterator[U]:
        return iter(self.elements)

    def op_intersection[T: PlainSet](self: T, other: U | T) -> T:
        return type(self)(*(self.elements & type(self).from_value(other).elements))

    def op_union[T: PlainSet](self: T, other: U | T) -> T:
        return type(self)(*(self.elements | type(self).from_value(other).elements))

    def is_subset_of[T: PlainSet](self: T, other: U | T) -> bool:
        try:
            other_set = type(self).from_value(other)
        except Exception:
            return False
        return self.elements.issubset(other_set.elements)

    def is_single_element(self) -> bool:
        return len(self.elements) == 1

    @classmethod
    def from_value[T: PlainSet](cls: type[T], value: Any) -> T:
        if isinstance(value, cls):
            return value
        return cls(value)

    # operators

    def __and__[T: PlainSet](self: T, other: U | T) -> T:
        return self.op_intersection(other)

    def __or__[T: PlainSet](self: T, other: U | T) -> T:
        return self.op_union(other)

    def __eq__(self, value: Any) -> bool:
        try:
            other = type(self).from_value(value)
        # TODO thats a bit whack
        except Exception:
            return False
        if not isinstance(other, PlainSet):
            return False
        return self.elements == other.elements

    def __contains__(self, item: U) -> bool:
        return item in self.elements

    def any(self) -> U:
        if self.is_empty():
            raise ValueError("no elements in set")
        return next(iter(self.elements))

    def is_finite(self) -> bool:
        return True


type BoolSetLike_ = bool | BoolSet | PlainSet[bool]


class BoolSet(PlainSet[bool]):
    # TODO rethink this
    @override
    def __contains__(self, item: BoolSetLike_) -> bool:
        return all(o_v in self.elements for o_v in BoolSet.from_value(item).elements)

    def op_not(self) -> "BoolSet":
        return BoolSet(*(not v for v in self.elements))

    def op_and(self, other: BoolSetLike_) -> "BoolSet":
        return BoolSet(
            *[
                v and o_v
                for v in self.elements
                for o_v in BoolSet.from_value(other).elements
            ]
        )

    def op_or(self, other: BoolSetLike_) -> "BoolSet":
        return BoolSet(
            *[
                v or o_v
                for v in self.elements
                for o_v in BoolSet.from_value(other).elements
            ]
        )

    @classmethod
    def unbounded(cls) -> "BoolSet":
        return cls(True, False)

    @override
    def serialize_pset(self) -> dict:
        return {"elements": list(self.elements)}

    @override
    @classmethod
    def deserialize_pset(cls, data: dict):
        return cls(*data["elements"])


BoolSetLike = bool | BoolSet | PlainSet[bool]


class EnumSet[E: Enum](PlainSet[SerializableEnum.Value[E]]):
    def __init__(
        self,
        *elements: "E | EnumSet[E] | type[E] | SerializableEnum.Value[E] | SerializableEnum",  # noqa: E501
    ):
        enum_types = (
            {e.enum.enum for e in elements if isinstance(e, EnumSet)}
            | {type(e) for e in elements if isinstance(e, Enum)}
            | {e for e in elements if isinstance(e, type) and issubclass(e, Enum)}
            | {e._enum.enum for e in elements if isinstance(e, SerializableEnum.Value)}
        )
        assert len(enum_types) == 1
        self.enum = SerializableEnum(next(iter(enum_types)))

        elements_raw = [
            *[elem for e in elements if isinstance(e, EnumSet) for elem in e.elements],
            *[e for e in elements if isinstance(e, self.enum.enum)],
            *[e for e in elements if isinstance(e, SerializableEnum.Value)],
        ]
        elements_ser = [self.enum.make_value(e) for e in elements_raw]

        super().__init__(*elements_ser)

    def __contains__(self, item: E) -> bool:
        return super().__contains__(self.enum.make_value(item))

    @classmethod
    def empty(cls, enum: type[E]) -> "EnumSet[E]":
        return cls(enum)

    @classmethod
    def unbounded(cls, enum: type[E]) -> "EnumSet[E]":
        return cls(*enum)

    @override
    def serialize_pset(self) -> dict:
        return {
            "elements": [
                e.serialize() for e in sorted(self.elements, key=lambda e: e.name)
            ],
            "enum": self.enum.serialize(),
        }

    @override
    @classmethod
    def deserialize_pset(cls, data: dict):
        enum = SerializableEnum.deserialize(data["enum"])
        return cls(*(enum.deserialize_value(e) for e in data["elements"]), enum)


# Helpers ------------------------------------------------------------------------------
@overload
def as_lit(x: BoolSetLike) -> BoolSet: ...


@overload
def as_lit(x: "QuantitySetLike") -> "Quantity_Interval_Disjoint": ...


@overload
def as_lit(x: Enum) -> EnumSet: ...


def as_lit(x):
    return P_Set.from_value(x)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import itertools
import logging
import shutil
import tempfile
import zipfile
from pathlib import Path

import pathspec
import pathvalidate
from ruamel.yaml import YAML

import atopile.config
from atopile import version
from atopile.errors import UserBadParameterError
from faebryk.libs.util import not_none, once

logger = logging.getLogger(__name__)


def _get_non_excluded_project_files(cfg: atopile.config.ProjectConfig) -> list[Path]:
    """
    Get all files in the project that aren't excluded by .gitignore.

    Args:
        cfg: The project configuration

    Returns:
        A list of Path objects for all non-excluded files
    """
    from git import Repo

    prjroot = cfg.paths.root
    repo = Repo(search_parent_directories=True)

    # For gitignore patterns, we need to get all files and filter out the matched ones,
    # since gitignore patterns specify which files to exclude
    ignore_pattern_lines = [".git/"]
    for ignore_file in itertools.chain(
        # TODO: only search directories between cwd and root
        # TODO include other gitignore sources
        Path(repo.working_tree_dir or prjroot).glob("*.gitignore"),
        Path(repo.working_tree_dir or prjroot).glob("*.atoignore"),
    ):
        if not ignore_file.is_file():
            continue
        with open(ignore_file, "r") as f:
            ignore_pattern_lines.extend(line.strip() for line in f.readlines())

    ignore_spec = pathspec.GitIgnoreSpec.from_lines(ignore_pattern_lines)

    return [prjroot / f for f in ignore_spec.match_tree_files(prjroot, negate=True)]


class DistValidationError(Exception): ...


class Dist:
    def __init__(self, path: Path):
        self.path = path
        self.validate()

    def validate(self) -> None:
        # validate
        try:
            self.manifest
        except Exception as e:
            raise DistValidationError(f"Invalid dist, can't load manifest: {e}") from e
        if self.manifest.package is None:
            raise DistValidationError("Invalid dist, manifest has no package")
        if self.manifest.package.version is None:
            raise DistValidationError("Invalid dist, manifest has no package version")
        if self.manifest.package.identifier is None:
            raise DistValidationError(
                "Invalid dist, manifest has no package identifier"
            )

    @property
    @once
    def manifest(self) -> atopile.config.ProjectConfig:
        with zipfile.ZipFile(self.path, "r") as zip_file:
            manifest_file = zip_file.read(str(atopile.config.PROJECT_CONFIG_FILENAME))
            return atopile.config.ProjectConfig.from_bytes(manifest_file)

    @property
    def version(self) -> str:
        return not_none(self.manifest.package).version

    @property
    def identifier(self) -> str:
        return not_none(self.manifest.package).identifier

    @property
    def compiler_version_spec(self) -> str:
        return not_none(self.manifest.requires_atopile)

    @property
    def bytes(self) -> int:
        return self.path.stat().st_size

    @staticmethod
    def get_package_filename(cfg: atopile.config.ProjectConfig) -> str:
        if cfg.package is None:
            raise DistValidationError("Project has no package configuration")
        return (
            pathvalidate.sanitize_filename(
                f"{cfg.package.identifier}-{cfg.package.version}".replace("/", "-")
            )
            + ".zip"
        )

    @staticmethod
    def build_dist(
        cfg: atopile.config.ProjectConfig | Path,
        output_path: Path,
    ) -> "Dist":
        """
        Build a distribution package for the given project.
        """
        if isinstance(cfg, Path):
            # TODO better error handling

            try:
                _cfg = atopile.config.ProjectConfig.from_path(cfg)
            except Exception as e:
                raise DistValidationError(
                    f"Could not load project config at: {cfg}: {e}"
                ) from e
            if _cfg is None:
                raise DistValidationError(f"Could not load project config at: {cfg}")
            cfg = _cfg

        if cfg.package is None:
            raise DistValidationError("Project has no package configuration")

        package_filename = Dist.get_package_filename(cfg)
        matched_files = _get_non_excluded_project_files(cfg)
        # TODO: make this ./dist or the likes?
        with tempfile.TemporaryDirectory() as temp_dir:
            temp_path = Path(temp_dir)
            zip_path = temp_path / package_filename

            # Create the package config
            package_config_path = temp_path / atopile.config.PROJECT_CONFIG_FILENAME

            yaml = YAML(typ="rt")  # round-trip
            with (cfg.paths.root / atopile.config.PROJECT_CONFIG_FILENAME).open(
                "r", encoding="utf-8"
            ) as file:
                config_data: dict = yaml.load(file) or {}

            config_data["package"]["identifier"] = str(cfg.package.identifier)
            config_data["package"]["repository"] = str(cfg.package.repository)
            config_data["package"]["version"] = str(cfg.package.version)

            with package_config_path.open("w", encoding="utf-8") as file:
                yaml.dump(config_data, file)

            ## Validate the package config is a valid config at all
            try:
                atopile.config.ProjectConfig.from_path(package_config_path)
            except Exception as e:
                raise UserBadParameterError(
                    "Something went wrong while validating the package config. "
                    "Please check the config file."
                ) from e

            # Bundle up the package
            with zipfile.ZipFile(
                zip_path, "x", compression=zipfile.ZIP_BZIP2, compresslevel=9
            ) as zip_file:
                ## Copy in the freshly minted package config
                zip_file.write(
                    package_config_path, atopile.config.PROJECT_CONFIG_FILENAME
                )

                ## Copy in the files to package
                for file in matched_files:
                    # rich.progress.track(
                    #    matched_files, description=f"Building {package_filename}..."
                    # ):
                    src_path = cfg.paths.root / file
                    if not src_path.is_file():
                        continue

                    if file.name == "ato.yaml":
                        continue

                    zip_file.write(src_path, file.relative_to(cfg.paths.root))

            out_file = output_path / package_filename
            if out_file.exists():
                out_file.unlink()
            out_file.parent.mkdir(parents=True, exist_ok=True)
            shutil.move(zip_path, out_file)

            return Dist(out_file)

    def install(self, path: Path):
        if path.exists():
            raise FileExistsError(f"Path {path} already exists")

        if not version.match(
            self.compiler_version_spec,
            version.get_installed_atopile_version(),
        ):
            raise version.VersionMismatchError(
                f"Compiler version {version.get_installed_atopile_version()} "
                f"does not match required version {self.compiler_version_spec}"
                f"of package {self.identifier}@{self.version}"
            )

        with zipfile.ZipFile(self.path, "r") as zip_file:
            zip_file.extractall(path)



"""CLI command definition for `ato build`."""

import logging
from typing import Annotated

import typer
from more_itertools import first

from atopile.cli.logging_ import NOW
from atopile.config import config
from atopile.telemetry import capture

logger = logging.getLogger(__name__)


@capture("cli:build_start", "cli:build_end")
def build(
    entry: Annotated[
        str | None,
        typer.Argument(
            help="Path to the project directory or build target address "
            '("path_to.ato:Module")'
        ),
    ] = None,
    selected_builds: Annotated[
        list[str], typer.Option("--build", "-b", envvar="ATO_BUILD")
    ] = [],
    target: Annotated[
        list[str], typer.Option("--target", "-t", envvar="ATO_TARGET")
    ] = [],
    frozen: Annotated[
        bool | None,
        typer.Option(
            help="PCB must be rebuilt without changes. Useful in CI",
            envvar="ATO_FROZEN",
        ),
    ] = None,
    keep_picked_parts: bool | None = None,
    keep_net_names: bool | None = None,
    keep_designators: bool | None = None,
    standalone: bool = False,
    open_layout: Annotated[
        bool | None, typer.Option("--open", envvar="ATO_OPEN_LAYOUT")
    ] = None,
):
    """
    Build the specified --target(s) or the targets specified by the build config.
    Optionally specify a different entrypoint with the argument ENTRY.
    eg. `ato build --target my_target path/to/source.ato:module.path`
    """
    from atopile import build as buildlib
    from atopile import buildutil
    from atopile.cli.install import check_missing_deps_or_offer_to_install
    from faebryk.libs.exceptions import accumulate, log_user_errors

    config.apply_options(
        entry=entry,
        selected_builds=selected_builds,
        target=target,
        standalone=standalone,
        frozen=frozen,
        keep_picked_parts=keep_picked_parts,
        keep_net_names=keep_net_names,
        keep_designators=keep_designators,
    )

    check_missing_deps_or_offer_to_install()

    if open_layout is not None:
        config.project.open_layout_on_build = open_layout

    logger.info("Saving logs to %s", config.project.paths.logs / NOW)
    with accumulate() as accumulator:
        for build in config.builds:
            with accumulator.collect(), log_user_errors(logger), build:
                logger.info("Building '%s'", config.build.name)
                app = buildlib.init_app()

                # TODO: add a way to override the following with custom build machinery
                buildutil.build(app)

    logger.info("Build successful! 🚀")

    if config.should_open_layout_on_build():
        selected_build_names = list(config.selected_builds)
        if len(selected_build_names) == 1:
            build = config.project.builds[first(selected_build_names)]
            try:
                from faebryk.libs.app.pcb import open_pcb

                open_pcb(build.paths.layout)
            except FileNotFoundError:
                pass
            except RuntimeError as e:
                logger.info(
                    f"{e.args[0]}\nReload pcb manually by pressing Ctrl+O; Enter"
                )

        elif len(selected_build_names) > 1:
            logger.warning(
                "`--open` option is only supported when building "
                "a single build. It will be ignored."
            )
